import yaml

DOCUMENTATION_URL = "https://invanalabs.github.io/invana-bot/spiders/manifest/"


class InvanaBotManifestItemBase(object):
    REQUIRED_FIELDS = []
    VALIDATOR_NAME = None

    def __init__(self, data=None):
        self.validate(data=data)

    def setup(self, data=None):
        raise NotImplementedError()

    def _validate_required_keys(self, data=None):
        data_keys = data.keys()
        errors = []
        for required_field in self.REQUIRED_FIELDS:
            if required_field not in data_keys:
                errors.append("{} key not defined in the {}".format(required_field, self.VALIDATOR_NAME))
        return errors

    def validate(self, data=None):
        errors = self._validate_required_keys(data=data)
        if errors.__len__() == 0:
            self.setup(data=data)
        return self, errors


class ManifestValidator(object):
    """

    """

    REQUIRED_KEYS = ("cti_id", "init_spider", "spiders", "settings", "context")

    def validate_manifest(self, manifest=None):
        """

        :param manifest: in json format.
        :return: (Manifest instance, errors)
        """
        errors = []
        manifest_keys = manifest.items()
        for required_key in self.REQUIRED_KEYS:
            if required_key not in manifest_keys:
                errors.append("{} key not defined in the manifest")

        self.validate_init_spider(data=manifest.get("init_spider", None))

        for spider in manifest.get("init_spider", []):
            pass
        return self, errors

    def validate_init_spider(self, data=None):
        errors = []
        if data is None:
            errors.append("init_spider cannot be None. Refer documentation at {}".format(DOCUMENTATION_URL))

        return InvanaBotInitSpider(data=data), errors

    def validate_spiders(self):
        return self.settings

    def validate_settings(self):
        return self.context

    def validate_context(self):
        return self.context


class InvanaBotInitSpider(InvanaBotManifestItemBase):
    spider_id = None
    start_urls = []
    REQUIRED_FIELDS = ["spider_id", "start_urls"]
    VALIDATOR_NAME = "init_spider"

    def setup(self, data=None):
        self.spider_id = data.get("spider_id")
        self.start_urls = data.get("start_urls", [])


class InvanaBotSettings(object):
    allowed_domains = []
    download_delay = 0
    data_storage = []

    def get_data_storages(self):
        return self.data_storage


class InvanaBotSpider(InvanaBotManifestItemBase):
    spider_id = None
    allowed_domains = None
    extractors = []
    traversals = []

    REQUIRED_FIELDS = ["spider_id", "allowed_domains", "extractors", "traversals"]
    VALIDATOR_NAME = "spider"

    def get_data_schema(self):
        # this method works only for CustomExtractor where data types are specified.
        pass


class InvanaBotTraversal(object):
    traversal_id = None
    selector_type = None
    selector_value = None
    max_pages = None
    next_spider_id = None


class InvanaBotExtractor(object):
    extractor_id = None
    extractor_type = None
    data_selectors = []


class InvanaBotDataSelector(object):
    selector_id = None
    selector = None
    selector_type = None
    selector_attribute = None
    data_type = None


class InvanaBotDataStorage(object):
    storage_id = None
    storage_type = None
    connection_uri = None
    database_name = None
    collection_name = None
    unique_key = None



def get_spider_from_list(spiders=None, spider_id=None):
    for spider in spiders:
        if spider.get("spider_id") == spider_id:
            return spider
    raise Exception("spider_id: {} not found in the spiders data".format(spider_id))



class ExtractorBase(object):
    """




    """

    def __init__(self, response=None, extractor=None, extractor_id=None, extractor_fn=None):
        """
        :param response: html response of the request
        :param extractor: extractor configuration in json; this is optional in most cases.
        :param extractor_fn: extractor python lambda; this is optional in most cases.
        :param extractor_id: the field with which data is stored in database.
        """
        if None in [response, extractor, extractor_id]:
            raise Exception("Invalid input to the extractor class, response, extractor and extractor_id are mandatory")
        self.response = response
        self.extractor = extractor
        self.extractor_id = extractor_id
        self.extractor_fn = extractor_fn



import sys
import os
import yaml


class CTIManifestManager(object):
    """



    """
    ib_functions = None
    required_files = ["manifest.yml", "ib_functions.py"]
    manifest = None

    def __init__(self, manifest_path=None):
        print("Setting ETI path as: {}".format(manifest_path))
        self.manifest_path = manifest_path

    def import_files(self):
        print("self.manifest_path", self.manifest_path)
        self.manifest = yaml.load(open("{}/manifest.yml".format(self.manifest_path)), Loader=yaml.FullLoader)
        sys.path.append(self.manifest_path)
        """
        don't remove the import below, this will be the cti_transformations.py,
        which is one of the required file to run the job. This file will be provided by the 
        user during the run.
        """
        try:
            import ib_functions
        except Exception as e:
            ib_functions = None
        self.ib_functions = ib_functions
        print ("self.ib_functions is {}".format(ib_functions))
        # print("manifest is {}".format(self.manifest))
        # print("ib_functions is {}".format(self.ib_functions))

    def validate_cti_path_and_files(self):
        errors = []

        try:
            files_in_path = os.listdir(self.manifest_path)
        except Exception as e:
            errors.append("No such path exist {}".format(self.manifest_path))
            files_in_path = []
        if errors == 0:
            for required_file in self.required_files:
                if required_file not in files_in_path:
                    errors.append("{} file not in the path {}".format(required_file, self.manifest_path))
        return errors

    def import_cti_transformations(self):
        if self.ib_functions:
            for transformation in self.manifest.get("transformations", []):
                method_to_call = getattr(self.ib_functions, transformation.get("transformation_fn"))
                transformation['transformation_fn'] = method_to_call

    def import_extractor_functions(self):
        if self.ib_functions:
            for spider in self.manifest.get("spiders", []):
                for extractor in spider.get("extractors", []):
                    method_to_call = getattr(self.ib_functions, extractor.get("extractor_fn"))
                    extractor['extractor_fn'] = method_to_call

    def get_manifest(self):
        errors = self.validate_cti_path_and_files()
        if len(errors) > 0:
            return None, errors
        self.import_files()
        self.import_cti_transformations()
        self.import_extractor_functions()
        print ("=====+++++++++============")
        return self.manifest, errors



DEFAULT_SETTINGS_BASE = {
    'COMPRESSION_ENABLED': False,  # this will save the data in normal text form, otherwise to bytes form.
    'HTTPCACHE_ENABLED': True,
    'TELNETCONSOLE_PORT': [6023, 6073],
    'ITEM_PIPELINES': {
        'invana_bot.core.storages.default.InvanaDataPipeline': 1,
        # 'invana_bot.core.storages.elasticsearch.ElasticSearchPipeline': 1,
        # 'invana_bot.core.storages.mongodb.MongoDBPipeline': 1
    },
    # 'INVANA_BOT_SETTINGS': {
    #     'HTTPCACHE_STORAGE_SETTINGS': {
    #         'CONNECTION_URI': "127.0.0.1",
    #         'DATABASE_NAME': "spider_cache_db",
    #         'COLLECTION_NAME': "web_link",
    #         "EXPIRY_TIME": 3600
    #     },
    #     'ITEM_PIPELINES_SETTINGS': {
    #         'CONNECTION_URI': "127.0.0.1",
    #         'DATABASE_NAME': "spider_data",
    #         'COLLECTION_NAME': "website_data"
    #     }
    # }

}

# MONGODB_SETTINGS = {
#     'ITEM_PIPELINES': {'invana_bot.core.storages.mongodb.MongoDBPipeline': 1},
#     # 'HTTPCACHE_STORAGE': "invana_bot.core.httpcache.mongodb.MongoDBCacheStorage",
#
#     'INVANA_BOT_SETTINGS': {
#         'HTTPCACHE_STORAGE_SETTINGS': {
#             'CONNECTION_URI': "127.0.0.1",
#             'DATABASE_NAME': "spider_cache_db",
#             'COLLECTION_NAME': "web_link",
#             "EXPIRY_TIME": 3600
#         },
#         'ITEM_PIPELINES_SETTINGS': {
#             'CONNECTION_URI': "127.0.0.1",
#             'DATABASE_NAME': "spider_data",
#             'COLLECTION_NAME': "website_data"
#         }
#     }
# }
#
# ELASTICSEARCH_SETTINGS = {
#     'ITEM_PIPELINES': {'invana_bot.core.storages.elasticsearch.ElasticSearchPipeline': 1},
#     # 'HTTPCACHE_STORAGE': "invana_bot.core.httpcache.elasticsearch.ESCacheStorage",
#
#     'INVANA_BOT_SETTINGS': {
#         'HTTPCACHE_STORAGE_SETTINGS': {
#             'CONNECTION_URI': "127.0.0.1",
#             'DATABASE_NAME': "spider_cache_db",
#             'COLLECTION_NAME': "web_link",
#             "EXPIRY_TIME": 3600
#         },
#         'ITEM_PIPELINES_SETTINGS': {
#             'CONNECTION_URI': "127.0.0.1",
#             'DATABASE_NAME': "spider_data",
#             'COLLECTION_NAME': "website_data"
#         }
#     }
# }

DEFAULT_SETTINGS = {}
DEFAULT_SETTINGS.update(DEFAULT_SETTINGS_BASE)
# DEFAULT_SETTINGS.update(MONGODB_SETTINGS)



from .base import InvanaBotJobGeneratorBase
from invana_bot.core.engines.cti import InvanaBotRunnerEngine
from datetime import datetime


class InvanaBotJobGenerator(InvanaBotJobGeneratorBase):
    """
        Invana-bot cti job generator
    """

    # def create_job(self,
    #                spider_config=None,
    #                context=None,
    #                spider_cls=None):
    #     if context is None:
    #         context = {}
    #     if 'job_id' not in context.keys():
    #         context['job_id'] = self.job_id
    #         context['job_started'] = datetime.now()
    #
    #     settings_from_manifest = spider_config.get("settings", {})
    #     actual_settings = self.settings
    #     actual_settings['DOWNLOAD_DELAY'] = settings_from_manifest.get("download_delay", 0)
    #     runner = CrawlerRunnerEngineBase(settings=actual_settings,
    #                                  spider_config=spider_config,
    #                                  job_id=self.job_id,
    #                                  context=context,
    #                                  spider_cls=spider_cls)
    #     job, errors = runner.crawl()
    #     return {"spider_job": job, "spider_job_errors": errors, "runner": runner}
    #

    def create_job(self, manifest=None, context=None, spider_cls=None, extra_arguments=None):
        if context is None:
            context = {}
        if 'job_id' not in context.keys():
            context['job_id'] = self.job_id
            context['job_started'] = datetime.now()

        settings_from_manifest = manifest.get("settings", {})
        print("============")
        print("============")
        print(settings_from_manifest)

        actual_settings = self.settings
        for k, v in settings_from_manifest.items():
            actual_settings[k.upper()] = v
        print(actual_settings)
        print("============")
        print("============")

        # actual_settings['DOWNLOAD_DELAY'] = settings_from_manifest.get("download_delay", 0)
        # actual_settings['ALLOWED_DOMAINS'] = settings_from_manifest.get("allowed_domains", [])
        runner = InvanaBotRunnerEngine(
            job_id=self.job_id,
            spider_cls=spider_cls,
            manifest=manifest,

            # settings=actual_settings,
            # context=context,
            # extra_arguments=extra_arguments
        )
        job, errors = runner.crawl()
        print("<<<<<<<<<<<<<<<<<<<<<<<=======", )
        print("<<<<<<<<<<<<<<<<<<<<<<<=======", )
        print("job errors", errors)
        print("<<<<<<<<<<<<<<<<<<<<<<<=======", )
        print("<<<<<<<<<<<<<<<<<<<<<<<=======", )
        job["spider_settings"] = actual_settings
        # job["spider_kwargs"]["default_storage"] = self.get_default_storage(settings=actual_settings)
        return {"spider_job": job, "spider_job_errors": errors, "runner": runner}



import locale
import re


class FieldTransformerBase(object):
    """

    IntField(data="1").transform()



    """

    def get_method(self):
        """
        Should return the function that takes data as input. Ex: str, int, float,

        :return:
        """
        raise NotImplementedError("This method is not implemented")

    def __init__(self, data=None):
        self.data = data

    def try_or_none(self):
        transformation_method = self.get_method()
        try:
            result_data = transformation_method(self.data)
        except Exception as e:
            print(e)
            result_data = self.data
        return result_data

    def transform(self):
        return self.try_or_none()


class StringField(FieldTransformerBase):

    def get_method(self):
        return str


class IntField(FieldTransformerBase):
    def get_method(self):
        def custom_int(data):
            data = locale.atoi(data)
            return int(data)

        return custom_int


class FloatField(FieldTransformerBase):
    def get_method(self):
        def custom_float(data):
            data = re.findall(r"[-+]?\d*\.\d+|[-+]?\d+", data)
            if len(data) > 0:
                return float(data[0])
            else:
                return float(0)
        return custom_float


class DictField(FieldTransformerBase):
    def get_method(self):
        return dict


class RawField(FieldTransformerBase):
    def get_method(self):
        return None



from scrapy.spiders import XMLFeedSpider
from invana_bot.utils.url import get_domain, get_absolute_url
from scrapy.utils.spider import iterate_spider_output
from invana_bot.utils.spiders import get_spider_from_list
import scrapy


class GenericXMLFeedSpider(XMLFeedSpider):
    """

    start_urls: ["https://news.ycombinator.com/rss"]
    allowed_domains: ["news.ycombinator.com"]
    itertag = "item"



    NOTE: paginating parameter

    wordpress: ?paged=n
    blogger: ?start-index=n
    """
    name = "GenericXMLFeedSpider"

    @staticmethod
    def run_extractor(node=None, response=None, extractor=None):
        item = {}
        for selector in extractor.get("data_selectors", []):
            data_type = selector.get("data_type").lower()
            selector_attribute = selector.get("selector_attribute")
            selector_id = selector.get("selector_id")
            selector = selector.get("selector")
            if data_type.startswith("List"):
                try:
                    item[selector_id] = node.xpath('{}/{}'.format(selector, selector_attribute), ).extract()
                except Exception as e:
                    print(e)
                    item[selector_id] = None
            else:
                try:
                    item[selector_id] = node.xpath('{}/{}'.format(selector, selector_attribute), ).extract_first()
                except Exception as e:
                    print(e)
                    item[selector_id] = None
        return item

    def parse_error(self, failure):
        """
        https://stackoverflow.com/questions/31146046/how-do-i-catch-errors-with-scrapy-so-i-can-do-something-when-i-get-user-timeout
        :return:
        """
        pass

    def post_parse(self, response=None):
        pass

    def parse_node(self, response, node, extractor):
        return self.run_extractor(node=node, response=response, extractor=extractor)

    @staticmethod
    def is_this_request_from_same_traversal(response, traversal):
        """
        This mean the current request came from this  traversal,
        so we can put max pages condition on this, otherwise for different
        traversals of different spiders, adding max_page doest make sense.
        """
        traversal_id = traversal['traversal_id']
        current_request_traversal_id = response.meta.get('current_request_traversal_id', None)
        return current_request_traversal_id == traversal_id

    def parse_nodes(self, response, nodes):
        """This method is called for the nodes matching the provided tag name
        (itertag). Receives the response and an Selector for each node.
        Overriding this method is mandatory. Otherwise, you spider won't work.
        This method must return either a BaseItem, a Request, or a list
        containing any of them.
        """
        spider_config = response.meta.get("spider_config")
        spiders = response.meta.get("spiders")
        context = self.context or {}
        if None in [spider_config]:
            spider_config = self.spider_config
            spiders = self.spiders

        data = {"url": response.url, "domain": get_domain(response.url)}
        for extractor in spider_config.get('extractors', []):
            extracted_items = []
            for selector in nodes:
                ret = iterate_spider_output(self.parse_node(response, selector, extractor))
                for result_item in self.process_results(response, ret):
                    extracted_items.append(result_item)
            data[extractor['extractor_id']] = {}
            data[extractor['extractor_id']]['entries'] = extracted_items
        context["spider_id"] = spider_config.get("spider_id")
        data['context'] = context

        """
        if spider_traversal_id is None, it means this response originated from the 
        request raised by the start urls. 

        If it is Not None, the request/response is raised some traversal strategy.
        """
        current_request_traversal_id = response.meta.get('current_request_traversal_id', None)
        """
        In xml crawling current_request_traversal_page_count starts from 1, because there is no page 0.
        """
        current_request_traversal_page_count = response.meta.get('current_request_traversal_page_count', 1)

        """
        Note on current_request_spider_id:
        This can never be none, including the ones that are started by start_urls .
        """
        spider_config_id = spider_config.get("spider_id")

        spider_traversals = spider_config.get('traversals', [])
        for traversal in spider_traversals:
            next_spider_id = traversal['next_spider_id']
            iter_param = traversal['iter_param']

            next_spider = get_spider_from_list(spider_id=next_spider_id, spiders=spiders)

            traversal['allow_domains'] = next_spider.get("allowed_domains", [])
            traversal_id = traversal['traversal_id']
            traversal_max_pages = traversal.get('max_pages', 1)

            traversal_links = []
            is_this_request_from_same_traversal = self.is_this_request_from_same_traversal(response, traversal)
            print("is_this_request_from_same_traversal", is_this_request_from_same_traversal)
            print("current_request_traversal_page_count", current_request_traversal_page_count)
            print("traversal_max_pages", traversal_max_pages)
            print(" current_request_traversal_page_count <= traversal_max_pages",
                  current_request_traversal_page_count <= traversal_max_pages)
            shall_traverse = False

            if current_request_traversal_id is None:
                """
                start urls will not have this traversal_id set, so we should allow then to traverse
                """
                shall_traverse = True

            elif is_this_request_from_same_traversal and current_request_traversal_page_count <= traversal_max_pages:
                """
                This block will be valid for the traversals from same spider_id, ie., pagination of a spider 
                """

                shall_traverse = True

            elif is_this_request_from_same_traversal:
                """
                """
                shall_traverse = True

            elif is_this_request_from_same_traversal is False and current_request_traversal_page_count <= \
                    traversal_max_pages:
                """
                This for the spider_a traversing to spider_b, this is not pagination, but trsversing between 
                spiders.
                """
                shall_traverse = True
            print("shall_traverse: {}".format(traversal_id), shall_traverse)
            if shall_traverse:
                current_url = response.url
                clean_url_without_iter_param = current_url.split("?")[0] if "?" in current_url else current_url
                # this is already iterating, so ignore.
                print("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<", clean_url_without_iter_param)
                print("clean_url_without_iter_param", clean_url_without_iter_param)
                traversal_link = "{}?{}={}".format(clean_url_without_iter_param, iter_param,
                                                   current_request_traversal_page_count + 1)

                print("traversal_link", traversal_link)

                data[traversal_id] = {"traversal_urls": [traversal_link]}
                """
                Then validate for max_pages logic if traversal_id's traversal has any!.
                This is where the further traversal for this traversal_id  is decided 
                """
                max_pages = traversal.get("max_pages", 1)

                current_request_traversal_page_count += 1

                """
                we are already incrementing, the last number, so using <= might make it 6 pages when 
                max_pages is 5 
                """
                if current_request_traversal_page_count <= max_pages:
                    print("=======current_request_traversal_page_count", current_request_traversal_page_count)
                    print("-----------------------------------")
                    yield scrapy.Request(
                        traversal_link,
                        callback=self.parse,
                        errback=self.parse_error,
                        meta={
                            "spider_config": next_spider,
                            "spiders": spiders,
                            "current_request_traversal_id": traversal_id,
                            "current_request_traversal_page_count": current_request_traversal_page_count,

                        }
                    )

            print("=================================================")
            print("====traversal_links", traversal_id, len(traversal_links), traversal_links)
            print("=================================================")

        yield data

        self.post_parse(response=response)

# from scrapy.spiders import XMLFeedSpider, Spider
# from invana_bot.utils.url import get_domain
# import feedparser
#
#
# class GenericFeedSpider(XMLFeedSpider):
#     name = "generic_feed_spider"
#     itertag = 'item'
#     iterator = 'xml'  # This is actually unnecessary, since it's the default value
#
#     def get_or_none(self, selector):
#         try:
#             return selector.extract_first()
#         except Exception as e:
#             print(e)
#             return None
#
#     def parse_node(self, response, node):
#         title = self.get_or_none(node.select('title/text()'))
#
#         url = self.get_or_none(node.select('link/text()'))
#         description = self.get_or_none(node.select('description/text()'))
#         pub_date = self.get_or_none(node.select('pubDate/text()'))
#         category = self.get_or_none(node.select('category/text()'))
#         # image = node.select('item/media:content/url')
#
#         item = {}
#         item['title'] = title
#         item['url'] = url
#
#         item['pub_date'] = pub_date
#         item['category'] = category
#         item['description'] = description
#         item['domain'] = get_domain(response.url)
#         return item
#
#
# class RSSSpider(Spider):
#     """
#
#     https://gist.github.com/nyov/9c70e780ea80204559d6da5525228702
#
#     """
#     name = "rss"
#
#     def parse_feed(self, feed):
#         """ Parse RSS/Atom feed using feedparser
#         """
#         data = feedparser.parse(feed)
#         if data.bozo:
#             if (hasattr(data.bozo_exception, 'getLineNumber') and
#                     hasattr(data.bozo_exception, 'getMessage')):
#                 line = data.bozo_exception.getLineNumber()
#                 # segment = feed.split('\n')[line - 1]
#             # could still try to return data. not necessarily completely broken
#             return None
#         return data
#
#     def parse(self, response):
#         # parse downloaded content with feedparser (NOT re-downloading with feedparser)
#         feed = self.parse_feed(response.body)
#         if feed:
#             # grab some feed elements
#             # - https://pythonhosted.org/feedparser/common-rss-elements.html
#             # - https://pythonhosted.org/feedparser/common-atom-elements.html
#
#             # ns = feed.namespaces
#             # feed_title = feed.feed.title
#             # feed_link = feed.feed.link
#             # feed_desc = feed.feed.description
#
#             for entry in feed.entries:
#                 # have content?
#                 content = entry.get('content')
#                 if content:
#                     # content = content[0]
#                     content = content[0]['value']
#
#                 item = {
#                     # global feed data
#                     # 'feed_title': feed_title,
#                     # 'feed_link': feed_link,
#                     # 'feed_description': feed_desc,
#                     #
#                     # item entry data
#                     # 'url': response.url,
#                     'url': entry.link,
#                     'title': entry.title,
#                     'domain': get_domain(response.url),
#                     'description': entry.description,
#                     # 'date': entry.published,
#                     # 'date': entry.published_parsed,
#                     'pub_date': entry.updated_parsed,
#
#                     # optional
#                     'content': content,
#                     'type': entry.get('dc_type'),
#                 }
#
#                 yield item



from invana_bot.extractors.base import ExtractorBase
from invana_bot.utils.selectors import get_selector_element
from invana_bot.utils.url import get_urn, get_domain
import json


class ParagraphsExtractor(ExtractorBase):
    def run(self):
        data = {}
        extracted_data = []
        elements = self.response.css("p::text").extract()
        for el in elements:
            extracted_data.append(el)
        data[self.extractor_id] = [d.strip() for d in extracted_data]
        return data


class HeadingsExtractor(ExtractorBase):
    # TODO - implement this
    def run(self):
        data = {}
        extracted_data = []
        heading_tags = ["h1", "h2", "h3", "h4", "h5", "h6"]
        elements = self.response.css(",".join(heading_tags)).extract()
        for el in elements:
            extracted_data.append(el)
        data[self.extractor_id] = [d.strip() for d in extracted_data]
        return data


class MainContentExtractor(ExtractorBase):
    # TODO - implement this
    pass


class TableContentExtractor(ExtractorBase):
    def run(self):
        data = {}
        tables = []
        for table in self.response.css("table"):
            table_data = []
            table_headers = [th.extract() for th in table.css("thead tr th::text")]
            for row in table.css("tbody tr"):
                row_data = [td.extract() for td in row.css("td::text")]
                row_dict = dict(zip(table_headers, row_data))
                table_data.append(row_dict)
            tables.append(table_data)
        data[self.extractor_id] = tables
        return data


class MetaTagExtractor(ExtractorBase):
    def run(self):
        data = {}
        meta_data_dict = {}

        elements = self.response.css('meta')
        for element in elements:
            # for open graph type of meta tags
            meta_property = element.xpath("@{0}".format('property')).extract_first()
            if meta_property:
                meta_property = meta_property.replace(":", "__").replace(".", "__")
                meta_data_dict[meta_property] = element.xpath(
                    "@{0}".format('content')).extract_first() or element.xpath("@{0}".format('value')).extract_first()

            meta_name = element.xpath("@{0}".format('name')).extract_first()
            if meta_name:
                meta_name = meta_name.replace(":", "__").replace(".", "__")
                meta_data_dict["meta__{}".format(meta_name)] = element.xpath(
                    "@{0}".format('content')).extract_first() or element.xpath("@{0}".format('value')).extract_first()

        try:
            title = self.response.css('title::text').get()
            if title:
                meta_data_dict["title"] = title
        except Exception as e:
            pass
        data[self.extractor_id] = meta_data_dict
        return data


class CustomContentExtractor(ExtractorBase):

    def run(self):
        data = {}
        extracted_data = {}
        for selector in self.extractor.get('data_selectors', []):
            if selector.get('selector_attribute') == 'element' and len(selector.get('child_selectors', [])) > 0:
                # TODO - currently only support multiple elements strategy. what if multiple=False
                elements = self.response.css(selector.get('selector'))
                elements_data = []
                for el in elements:
                    datum = {}
                    for child_selector in selector.get('child_selectors', []):
                        _d = get_selector_element(el, child_selector)
                        datum[child_selector.get('selector_id')] = _d if _d else None
                    elements_data.append(datum)
                data_type = selector.get("data_type", "RawField")
                if data_type.startswith("List") is False:
                    single_data = elements_data[0]
                    extracted_data[selector.get('selector_id')] = single_data
                else:
                    extracted_data[selector.get('selector_id')] = elements_data
            else:
                _d = get_selector_element(self.response, selector)
                extracted_data[selector.get('selector_id')] = _d
        data[self.extractor_id] = extracted_data
        return data


class IconsExtractor(ExtractorBase):
    def run(self):
        data = {}
        meta_data_dict = {}

        favicon = self.response.xpath('//link[@rel="shortcut icon"]').xpath("@href").get()
        if favicon:
            meta_data_dict['favicon'] = favicon

        elements = self.response.xpath('//link[@rel="icon" or @rel="apple-touch-icon-precomposed"]')
        for element in elements:
            # for open graph type of meta tags
            meta_property = element.xpath("@sizes").extract_first()
            if meta_property:
                meta_property = meta_property.replace(":", "__").replace(".", "__")
                meta_data_dict[meta_property] = element.xpath("@{0}".format('href')).extract_first()
        data[self.extractor_id] = meta_data_dict
        return data


class JSONLDExtractor(ExtractorBase):

    def run(self):
        data = {}
        extracted_data = []

        elements = self.response.xpath('//script[@type="application/ld+json"]/text()').extract()
        for element in elements:
            # for open graph type of meta tags
            try:
                element = element.strip()
                element = json.loads(element)
                extracted_data.append(element)
            except Exception as e:
                pass

        data[self.extractor_id] = extracted_data
        return data


class PlainHTMLContentExtractor(ExtractorBase):

    def run(self):
        data = {}
        response_text = self.response.body
        data[self.extractor_id] = response_text
        return data


class FeedUrlExtractor(ExtractorBase):
    def run(self):
        data = {}
        data[self.extractor_id] = {}
        data[self.extractor_id]['rss__xml'] = self.response.xpath('//link[@type="application/rss+xml"]').xpath(
            "@href").extract_first()
        data[self.extractor_id]['rss__atom'] = self.response.xpath('//link[@type="application/atom+xml"]').xpath(
            "@href").extract_first()
        return data


class PageOverviewExtractor(ExtractorBase):

    def run(self):
        data = {}

        meta_tags_data = MetaTagExtractor(
            response=self.response,
            extractor=self.extractor,
            extractor_id=self.extractor_id
        ).run().get(self.extractor_id, {})

        paragraphs_data = ParagraphsExtractor(
            response=self.response,
            extractor=self.extractor,
            extractor_id="paragraphs"
        ).run().get("paragraphs", {})
        # TODO - clean the data, that is extracted. ex:  meta_tags_data.get("title")
        extracted_data = {
            "title":
                meta_tags_data.get("title") or
                meta_tags_data.get("meta__title") or
                meta_tags_data.get("og__title") or
                meta_tags_data.get("fb__title") or
                meta_tags_data.get("meta__twitter__title"),
            "description":
                meta_tags_data.get("description") or
                meta_tags_data.get("meta__description") or
                meta_tags_data.get("og__description") or
                meta_tags_data.get("fb__description") or
                meta_tags_data.get("meta__twitter__description"),
            "image":
                meta_tags_data.get("image") or
                meta_tags_data.get("meta__image") or
                meta_tags_data.get("og__image") or
                meta_tags_data.get("fb__image") or
                meta_tags_data.get("meta__twitter__image"),
            "url":
                meta_tags_data.get("url") or
                meta_tags_data.get("meta__url") or
                meta_tags_data.get("og__url") or
                meta_tags_data.get("fb__url") or
                meta_tags_data.get("meta__twitter__url"),
            "page_type": meta_tags_data.get("og__type"),
            "keywords": meta_tags_data.get("meta__keywords"),
            "domain": get_domain(self.response.url),
            "first_paragraph": paragraphs_data[0] if len(paragraphs_data) > 0 else None,
            "shortlink_url": self.response.xpath('//link[@rel="shortlink"]').xpath("@href").extract_first(),
            "canonical_url": self.response.xpath('//link[@rel="canonical"]').xpath("@href").extract_first()
        }
        data[self.extractor_id] = extracted_data
        return data



from .base import WebCrawlerBase
from importlib import import_module
from invana_bot.utils.url import get_domain
import scrapy


class InvanaBotSingleWebCrawler(WebCrawlerBase):
    """
    This is generic spider
    """
    name = "InvanaBotSingleWebCrawler"

    def closed(self, reason):
        print("spider closed with payload:", reason, self.spider_config.get('cti_id'))

    @staticmethod
    def run_extractor(response=None, extractor=None):
        extractor_type = extractor.get("extractor_type")
        extractor_id = extractor.get("extractor_id")

        driver_klass_module = import_module(f'invana_bot.extractors')
        driver_klass = getattr(driver_klass_module, extractor_type)

        if extractor_type is None:
            return {}
        else:
            try:
                extractor_object = driver_klass(response=response,
                                                extractor=extractor,
                                                extractor_id=extractor_id)
                data = extractor_object.run()
                return data
            except Exception as e:
                print(e)
        return {}

    def parse(self, response=None):
        self.logger.info("======Parsing the url: {}".format(response.url))
        spider_config = response.meta.get("spider_config")
        default_storage = self.get_default_storage(
            settings=self.settings,
            spider_config=spider_config
        )

        data = {}
        for extractor in spider_config.get('extractors', []):
            extracted_data = self.run_extractor(response=response, extractor=extractor)
            data.update(extracted_data)

        context = self.manifest.get("context")
        if context is not None:
            data.update({"context": context})
        data['url'] = response.url
        data['domain'] = get_domain(response.url)
        data['context']['spider_id'] = spider_config['spider_id']
        traversal_data, to_traverse_links_list = self.run_traversals(spider_config=spider_config, response=response)
        data.update(traversal_data)

        # This will save the data
        yield self.prepare_data_for_yield(
            data=data,
            storage_id=default_storage.get("storage_id"),
            collection_name=default_storage.get("collection_name")
        )

        # This will initiate new traversals
        traversal_requests = self.make_traversal_requests(to_traverse_links_list=to_traverse_links_list)
        for traversal_request in traversal_requests:
            yield traversal_request
        self.post_parse(response=response)



from urllib.parse import urlparse


def get_urn(url):
    """
    convert https://blog.scrapinghub.com/page/6/ into blog.scrapinghub.com/page/6/
    :param url:
    :return:
    """
    if "://" in url:
        return url.split("://")[1]
    return url


def get_domain(url):
    url_parsed = urlparse(url)
    return url_parsed.netloc

    # if "://" in url:
    #     url = url.split("://")[1]
    # if "/" in url:
    #     url = url.split("/")[0]
    # return url.strip()


def get_absolute_url(url=None, origin_url=None):
    url = url.lstrip("/")

    url_parsed = urlparse(origin_url)
    scheme = url_parsed.scheme
    host = url_parsed.netloc
    if "://" in url:
        return url
    else:
        url = "{}://{}/{}".format(scheme, host, url)
        return url



from .content import ParagraphsExtractor, TableContentExtractor, MetaTagExtractor, \
    CustomContentExtractor, IconsExtractor, JSONLDExtractor, PlainHTMLContentExtractor, \
    PageOverviewExtractor, FeedUrlExtractor
from .images import ImagesExtractor
from .links import AllLinksExtractor, AllLinksAnalyticsExtractor
from .python import PythonBasedExtractor
